from typing import TYPE_CHECKING, Dict, Any, Hashable, Tuple, Union, Iterable

if TYPE_CHECKING:
    from . import FBMessengerChannel


class ExperimentalFlagsManager:

    DEFAULT_VALUES = {
        'proxy_links_by_facebook': True,  # True then links are proxied by facebook
        'send_link_with_description': False,  # Send link messages with descriptions
        'show_pending_threads': False,  # Show threads pending approval in the thread list
        'show_archived_threads': False,  # Show archived threads in the thread list
    }

    def __init__(self, channel: 'FBMessengerChannel'):
        self.config: Dict[str, Any] = ExperimentalFlagsManager.DEFAULT_VALUES.copy()
        self.config.update(channel.config.get('flags', dict()))

    def __call__(self, flag_key: str) -> Any:
        if flag_key not in self.config:
            raise ValueError("%s is not a valid experimental flag" % flag_key)
        return self.config[flag_key]


def get_value(source: Union[dict, Iterable], path: Tuple[Hashable, ...], default=None) -> Any:
    """
    Get value from a path of keys
    Args:
        source: Data source
        path: Path to get value from
        default: Default value if the value is not found

    Returns:
        The value found or the default value.
    """

    data = source
    stack = path
    while stack:
        key = stack[0]
        stack = stack[1:]
        try:
            data = data.__getitem__(key)
        except (AttributeError, KeyError, IndexError, TypeError):
            return default
    return data



import os
import pickle
from io import BytesIO
from typing import Optional, List, IO, Set, Dict, Any

import requests
import yaml
from fbchat import FBchatUserError, ThreadLocation, logging
from fbchat.models import Thread
from ehforwarderbot import EFBChannel, EFBChat, EFBMsg, EFBStatus, ChannelType, MsgType
from ehforwarderbot import utils as efb_utils
from ehforwarderbot.utils import extra
from ehforwarderbot.status import EFBMessageRemoval
from ehforwarderbot.exceptions import EFBException, EFBOperationNotSupported

from .__version__ import __version__ as version
from .efms_chat import EFMSChat
from .efms_client import EFMSClient
from .master_messages import MasterMessageManager
from .extra_functions import ExtraFunctionsManager
from . import utils as efms_utils
from .utils import ExperimentalFlagsManager


class FBMessengerChannel(EFBChannel):
    channel_name: str = "Facebook Messenger Slave"
    channel_emoji: str = "⚡️"
    channel_id = "blueset.fbmessenger"
    channel_type: ChannelType = ChannelType.Slave
    __version__: str = version

    client: EFMSClient = None
    config: Dict[str, Any] = None

    logger = logging.getLogger(channel_id)

    def __init__(self):
        session_path = os.path.join(efb_utils.get_data_path(FBMessengerChannel.channel_id), "session.pickle")
        try:
            data = pickle.load(open(session_path, 'rb'))
            self.client = EFMSClient(None, None, session_cookies=data)
            self.client.channel = self
            EFMSChat.cache[EFBChat.SELF_ID] = EFMSChat(self,
                                                       self.client.fetchThreadInfo(self.client.uid)[self.client.uid])
        except FileNotFoundError:
            raise EFBException("Session not found, please authorize your account.\n"
                               "To do so, run: efms-auth")
        except FBchatUserError as e:
            message = str(e) + "\n" + \
                      "You may need to re-authorize your account.\n" + \
                      "To do so, run: efms-auth"
            raise EFBException(message)

        self.load_config()
        self.flag: ExperimentalFlagsManager = ExperimentalFlagsManager(self)
        self.master_message: MasterMessageManager = MasterMessageManager(self)
        self.extra_functions: ExtraFunctionsManager = ExtraFunctionsManager(self)

        # Initialize list of chat from server
        self.get_chats()

        # Monkey patching
        Thread.__eq__ = lambda a, b: a.uid == b.uid

    def load_config(self):
        """
        Load configuration from path specified by the framework.

        Configuration file is in YAML format.
        """
        config_path = efb_utils.get_config_path(self.channel_id)
        if not os.path.exists(config_path):
            self.config: Dict[str, Any] = dict()
            return
        with open(config_path) as f:
            self.config: Dict[str, Any] = yaml.load(f)

    def get_chats(self) -> List[EFMSChat]:
        locations = (ThreadLocation.INBOX,)
        if self.flag('show_pending_threads'):
            locations += (ThreadLocation.PENDING, ThreadLocation.OTHER)
        if self.flag('show_archived_threads'):
            locations += (ThreadLocation.ARCHIVED,)
        chats = list(map(lambda graphql: EFMSChat(self, graph_ql_thread=graphql),
                         self.client.get_thread_list(location=locations)))
        loaded_chats = set(i.chat_uid for i in chats)
        for i in self.client.fetchAllUsers():
            if i.uid not in loaded_chats:
                chats.append(EFMSChat(self, thread=i))

        return chats

    def get_chat(self, chat_uid: str, member_uid: Optional[str] = None) -> EFMSChat:
        thread_id = member_uid or chat_uid
        if thread_id in EFMSChat.cache:
            chat = EFMSChat.cache[thread_id]
        else:
            # thread = self.client.fetchThreadInfo(thread_id)[str(thread_id)]
            # chat = EFMSChat(self, thread=thread)
            graph_ql = self.client.get_thread_info(thread_id)
            chat = EFMSChat(self, graph_ql_thread=graph_ql)
        if member_uid:
            chat.is_chat = False
            chat.group = self.get_chat(chat_uid)
        return chat

    def send_message(self, msg: EFBMsg) -> EFBMsg:
        return self.master_message.send_message(msg)

    def send_status(self, status: EFBStatus):
        if isinstance(status, EFBMessageRemoval):
            raise EFBOperationNotSupported("Messages cannot be removed in Facebook Messenger.")
        # Other status types go here
        raise EFBOperationNotSupported()

    def poll(self):
        self.client.listen()

    def stop_polling(self):
        self.client.listening = False

    def get_chat_picture(self, chat: EFBChat) -> IO[bytes]:
        self.logger.debug("Getting picture of chat %s", chat)
        photo_url = EFMSChat.cache[chat.chat_uid] and \
                    EFMSChat.cache[chat.chat_uid].vendor_specific.get('profile_picture_url')
        self.logger.debug("[%s] has photo_url from cache: %s", chat.chat_uid, photo_url)
        if not photo_url:
            thread = self.client.get_thread_info(chat.chat_uid)
            photo_url = efms_utils.get_value(thread, ('messaging_actor', 'big_image_src', 'uri'))
        self.logger.debug("[%s] has photo_url from GraphQL: %s", chat.chat_uid, photo_url)
        if not photo_url:
            thread = self.client.fetchThreadInfo(chat.chat_uid)[chat.chat_uid]
            photo_url = getattr(thread, 'photo', None)
        self.logger.debug("[%s] has photo_url from legacy API: %s", chat.chat_uid, photo_url)
        if not photo_url:
            raise EFBOperationNotSupported('This chat has no picture.')
        photo = BytesIO(requests.get(photo_url).content)
        photo.seek(0)
        return photo

    # Extra functions

    @extra(name="Show threads list",
           desc="Usage:\n"
                "    {function_name}")
    def threads_list(self, args: str) -> str:
        return self.extra_functions.threads_list(args)

    @extra(name="Search for users",
           desc="Show the first 10 results."
                "Usage:\n"
                "    {function_name} keyword")
    def search_users(self, args: str) -> str:
        return self.extra_functions.search_users(args)

    @extra(name="Search for groups",
           desc="Show the first 10 results."
                "Usage:\n"
                "    {function_name} keyword")
    def search_groups(self, args: str) -> str:
        return self.extra_functions.search_groups(args)

    @extra(name="Search for pages",
           desc="Show the first 10 results."
                "Usage:\n"
                "    {function_name} keyword")
    def search_pages(self, args: str) -> str:
        return self.extra_functions.search_pages(args)

    @extra(name="Search for threads",
           desc="Show the first 10 results."
                "Usage:\n"
                "    {function_name} keyword")
    def search_threads(self, args: str) -> str:
        return self.extra_functions.search_threads(args)



import logging
import time
import threading
import emoji
from typing import TYPE_CHECKING, Set

import re
from fbchat.models import Thread, Message, TypingStatus, ThreadType, Mention, EmojiSize, MessageReaction
from ehforwarderbot import EFBMsg, MsgType
from ehforwarderbot.message import EFBMsgLinkAttribute, EFBMsgStatusAttribute, ChatType
from ehforwarderbot.exceptions import EFBMessageTypeNotSupported

from .utils import get_value

if TYPE_CHECKING:
    from . import FBMessengerChannel


class MasterMessageManager:

    logger = logging.getLogger("MasterMessageManager")

    def __init__(self, channel: 'FBMessengerChannel'):
        self.channel = channel
        channel.supported_message_types: Set[MsgType] = {MsgType.Text, MsgType.Image, MsgType.Sticker,
                                                         MsgType.Audio, MsgType.File, MsgType.Video,
                                                         MsgType.Status, MsgType.Unsupported}
        self.client = channel.client
        self.flag = channel.flag

    def send_message(self, msg: EFBMsg) -> EFBMsg:
        self.logger.debug("Received message from master: %s", msg)

        try:
            target_msg_offset = 0
            prefix = ""

            mentions = []

            # Send message reaction
            if msg.target and msg.text.startswith('r`') and getattr(MessageReaction, msg.text[2:], None)\
                    and msg.uid.startswith("mid.$"):
                self.logger.debug("[%s] Message is a reaction to another message: %s", msg.uid, msg.text)
                msg_id = ".".join(msg.target.uid.split(".", 2)[:2])
                self.client.reactToMessage(msg_id, getattr(MessageReaction, msg.text[2:]))
                msg.uid = "__reaction__"
                return msg

            # Target message
            if msg.target:
                self.logger.debug("[%s] Message replying to another message: %s", msg.uid, msg.target)
                if msg.target.chat.chat_type == ChatType.Group:
                    target_msg_name = msg.target.author.chat_alias or msg.target.author.chat_name
                    prefix = '@%s "%s"\n' % (target_msg_name, msg.target.text or msg.target.type)
                    mentions.append(Mention(msg.target.author.chat_uid, 1, len(target_msg_name)))
                else:
                    prefix = '"%s"\n' % msg.target.text or msg.target.type
                target_msg_offset = len(prefix)
                self.logger.debug("[%s] Converted to prefix: %s", msg.uid, prefix)

            # Message substitutions
            if msg.substitutions:
                self.logger.debug("[%s] Message has substitutions: %s", msg.uid, msg.substitutions)
                for i in msg.substitutions:
                    mentions.append(Mention(msg.substitutions[i].chat_uid,
                                            target_msg_offset + i[0], i[1] - i[0]))
                self.logger.debug("[%s] Translated to mentions: %s", msg.uid, mentions)

            fb_msg = Message(text=prefix + msg.text, mentions=mentions)
            thread: Thread = self.client.fetchThreadInfo(msg.chat.chat_uid)[str(msg.chat.chat_uid)]

            if msg.type in (MsgType.Text, MsgType.Unsupported):
                if msg.text[:-1] in emoji.UNICODE_EMOJI and msg.text[-1] in 'SML':
                    self.logger.debug("[%s] Message is an Emoji message: %s", msg.uid, msg.text)
                    if msg.text[-1] == 'S':
                        fb_msg.emoji_size = EmojiSize.SMALL
                    elif msg.text[-1] == 'M':
                        fb_msg.emoji_size = EmojiSize.MEDIUM
                    elif msg.text[-1] == 'L':
                        fb_msg.emoji_size = EmojiSize.LARGE
                    fb_msg.text = msg.text[:-1]
                msg.uid = self.client.send(fb_msg, thread_id=thread.uid, thread_type=thread.type)
            elif msg.type in (MsgType.Image, MsgType.Sticker):
                msg.uid = self.client.send_image_file(msg.path, msg.mime, message=fb_msg,
                                                      thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Audio:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_audio(file_id=file_id, message=fb_msg,
                                                 thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.File:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_file(file_id=file_id, message=fb_msg,
                                                thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Video:
                file_id = self.upload_file(msg)
                msg.uid = self.client.send_video(file_id=file_id, message=fb_msg,
                                                 thread_id=thread.uid, thread_type=thread.type)
            elif msg.type == MsgType.Status:
                attribute: EFBMsgStatusAttribute = msg.attributes
                if attribute.status_type in (EFBMsgStatusAttribute.Types.TYPING,
                                             EFBMsgStatusAttribute.Types.UPLOADING_AUDIO,
                                             EFBMsgStatusAttribute.Types.UPLOADING_VIDEO,
                                             EFBMsgStatusAttribute.Types.UPLOADING_IMAGE,
                                             EFBMsgStatusAttribute.Types.UPLOADING_FILE):
                    self.client.setTypingStatus(TypingStatus.TYPING, thread_id=thread.uid, thread_type=thread.type)
                    threading.Thread(target=self.stop_typing, args=(attribute.timeout, thread.uid, thread.type)).run()
            elif msg.type == MsgType.Link:
                attribute: EFBMsgLinkAttribute = msg.attributes
                if self.flag('send_link_with_description'):
                    info = (attribute.title,)
                    if attribute.description:
                        info += (attribute.description,)
                    info += (attribute.url,)
                    text = "\n".join(info)
                else:
                    text = attribute.url
                if fb_msg.text:
                    text = fb_msg.text + "\n" + text
                fb_msg.text = text
                msg.uid = self.client.send(fb_msg, thread_id=thread.uid, thread_type=thread.type)
            else:
                raise EFBMessageTypeNotSupported()
            return msg
        finally:
            if msg.file and not msg.file.closed:
                msg.file.close()
            self.client.markAsSeen()
            self.client.markAsRead(msg.chat.chat_uid)

    def stop_typing(self, timeout: int, thread_uid: str, thread_type: ThreadType):
        """Wait for a number of milliseconds, and stop typing."""
        time.sleep(timeout / 1000)
        self.client.setTypingStatus(TypingStatus.STOPPED, thread_id=thread_uid, thread_type=thread_type)

    def upload_file(self, msg: EFBMsg) -> int:
        """Upload media of a message as file, and return the file id."""
        response = self.client._postFile(self.client.req_url.UPLOAD, {
            'file': (msg.filename, msg.file, msg.mime)
        }, fix_request=True, as_json=True)
        return get_value(response, ('payload', 'metadata', 0, 'file_id')) or \
               get_value(response, ('payload', 'metadata', 0, 'audio_id')) or \
               get_value(response, ('payload', 'metadata', 0, 'video_id'))



import copy
import os
import urllib.parse
import logging
from typing import TYPE_CHECKING, Optional, Any, Dict, Tuple
from tempfile import NamedTemporaryFile

import requests
from fbchat import Client, ThreadType, FBchatException, logging, ThreadLocation, GraphQL
from fbchat.models import Message, EmojiSize
from fbchat.utils import check_request, get_jsmods_require, ReqUrl
from ehforwarderbot import EFBMsg, MsgType, coordinator
from ehforwarderbot.message import EFBMsgLinkAttribute, EFBMsgLocationAttribute

from .efms_chat import EFMSChat
from .utils import get_value


if TYPE_CHECKING:
    from . import FBMessengerChannel


class EFMSClient(Client):
    channel: "FBMessengerChannel" = None

    logger = logging.getLogger("EFMSClient")

    """Set of messages IDs sent by EFMS, popped when message is received again."""
    sent_messages = set()

    # Overrides for patches

    def __init__(self, *args, **kwargs):
        kwargs['logging_level'] = logging.NOTSET
        super().__init__(*args, **kwargs)
        self.logger = logging.getLogger(__name__)

    def fetchImageUrl(self, image_id):
        """Fetches the url to the original image from an image attachment ID
        Overrides original method with a bug fix.

        Args:
            image_id (str): The image you want to fetch

        Returns:
            str: An url where you can download the original image

        Raises:
            FBChatException: if request failed
        """
        image_id = str(image_id)
        j = check_request(self._get(ReqUrl.ATTACHMENT_PHOTO, query={'photo_id': str(image_id)}))

        url = get_jsmods_require(j, 3)
        if url is None:
            raise FBchatException('Could not fetch image url from: {}'.format(j))
        return url

    def send_image_file(self, image_path, mimetype, message=None, thread_id=None, thread_type=ThreadType.USER):
        """
        Sends a local image to a thread

        Args:
            image_path: Path of an image to upload and send
            mimetype: MIME type of the image
            message: Additional message
            thread_id: User/Group ID to send to. See :ref:`intro_threads`
            thread_type (models.ThreadType): See :ref:`intro_threads`

        Returns:
            :ref:`Message ID <intro_message_ids>` of the sent image

        Raises:
            FBchatException: if request failed
        """
        thread_id, thread_type = self._getThread(thread_id, thread_type)
        is_gif = (mimetype == 'image/gif')
        image_id = self._uploadImage(image_path, open(image_path, 'rb'), mimetype)
        return self.sendImage(image_id=image_id, message=message, thread_id=thread_id,
                              thread_type=thread_type, is_gif=is_gif)

    def markAsDelivered(self, thread_id, message_id):
        """Mark message as delivered"""
        super().markAsDelivered(thread_id, message_id)

    # Original methods

    def get_thread_list(self, limit: int=50, before: int=None,
                        location: Tuple[ThreadLocation, ...]=(ThreadLocation.INBOX,)):
        location = list(i.name for i in location)

        j = self.graphql_request(GraphQL(doc_id='1349387578499440', params={
            'limit': limit,
            'tags': location,
            'includeDeliveryReceipts': True,
            'includeSeqID': False,
            'before': before
        }))

        if j.get('viewer') is None:
            raise FBchatException('Could not fetch thread list: {}'.format(j))

        return j['viewer']['message_threads']['nodes']

    def get_thread_info(self, id: str):
        j = self.graphql_request(GraphQL(doc_id='1508526735892416', params={
            'id': id,
            'message_limit': 0,
            'load_message': 0,
            'load_read_receipt': False,
            'before': None,
        }))

        if j.get('message_thread') is None:
            raise FBchatException('Could not fetch thread list: {}'.format(j))

        return j['message_thread']

    def process_url(self, url: str) -> str:
        """Unwrap Facebook-proxied URL if necessary."""
        if self.client.flag('proxy_links_by_facebook'):
            return url
        if 'safe_image.php' in url:
            query = urllib.parse.urlparse(url).query
            escaped = urllib.parse.parse_qs(query).get('url', [url])[0]
            return escaped
        elif url.startswith('https://l.facebook.com/l.php'):
            query = urllib.parse.urlparse(url).query
            escaped = urllib.parse.parse_qs(query).get('u', [url])[0]
            return escaped
        else:
            return url

    def attach_media(self, msg: EFBMsg, attachment: Dict[str, Any]):
        """
        Attach an attachment to a message.

        Args:
            msg: Message to be attached to
            attachment:
                Dict of information of the attachment
                ``fbchat`` entity is not used as it is not completed.
        """
        self.logger.debug("[%s] Trying to attach media: %s", msg.uid, attachment)

        blob_attachment: Dict[str, Any] = attachment.get('mercury', {}).get('blob_attachment', {})
        attachment_type: str = blob_attachment.get("__typename", None)
        if 'sticker_attachment' in attachment.get('mercury', {}):
            attachment_type = '__Sticker'
        if 'extensible_attachment' in attachment.get('mercury', {}):
            attachment_type = '__Link'
            if get_value(attachment, ('mercury', 'extensible_attachment',
                                      'story_attachment', 'target', '__typename')) == 'MessageLocation':
                attachment_type = 'MessageLocation'

        self.logger.debug("[%s] The attachment has type %s", msg.uid, attachment_type)

        msg.filename = attachment.get('filename', None)
        msg.mime = attachment.get('mimeType', None)
        if attachment_type == "MessageAudio":
            msg.type = MsgType.Audio
            msg.filename = msg.filename or 'audio.mp4'
            msg.mime = msg.mime or 'audio/mpeg'
            ext = os.path.splitext(msg.filename)[1]
            msg.file = NamedTemporaryFile(suffix=ext)
            msg.file.write(requests.get(blob_attachment['playable_url']).content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == 'MessageImage':
            msg.type = MsgType.Image
            msg.filename = msg.filename or 'image.png'
            msg.mime = msg.mime or 'image/png'
            attribution_app = get_value(blob_attachment, ('attribution_app', 'name'))
            if attribution_app:
                if msg.text:
                    msg.text += " (via %s)" % attribution_app
                else:
                    msg.text = "via %s" % attribution_app
            ext = os.path.splitext(msg.filename)[1]
            msg.file = NamedTemporaryFile(suffix=ext)
            url = self.fetchImageUrl(attachment['id'])
            msg.file.write(requests.get(url).content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == 'MessageAnimatedImage':
            msg.type = MsgType.Image
            msg.filename = msg.filename or 'image.gif'
            msg.mime = msg.mime or 'image/gif'
            ext = os.path.splitext(msg.filename)[1]
            msg.file = NamedTemporaryFile(suffix=ext)
            msg.file.write(requests.get(blob_attachment['animated_image']['uri']).content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == 'MessageFile':
            msg.type = MsgType.Image
            msg.filename = msg.filename or 'file'
            msg.mime = msg.mime or 'application/octet-stream'
            ext = os.path.splitext(msg.filename)[1]
            msg.file = NamedTemporaryFile(suffix=ext)
            msg.file.write(requests.get(blob_attachment['url']).content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == 'MessageVideo':
            msg.type = MsgType.Image
            msg.filename = msg.filename or 'video.mp4'
            msg.mime = msg.mime or 'video/mpeg'
            ext = os.path.splitext(msg.filename)[1]
            msg.file = NamedTemporaryFile(suffix=ext)
            msg.file.write(requests.get(blob_attachment['playable_url']).content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == '__Sticker':
            if get_value(attachment, ('mercury', 'sticker_attachment', 'pack', 'id')) == "227877430692340":
                self.logger.debug("[%s] Sticker received is a \"Like\" sticker. Converting message to text.", msg.uid)

                sticker_id = get_value(attachment, ('mercury', 'sticker_attachment', 'id'))
                for i in EmojiSize:
                    if sticker_id == i.value:
                        msg.type = MsgType.Text
                        msg.text = "👍 (%s)" % i.__name__[0]
                        return
            msg.type = MsgType.Sticker
            url = attachment['mercury']['sticker_attachment']['url']
            msg.text = attachment['mercury']['sticker_attachment'].get('label', '')
            msg.filename = os.path.split(urllib.parse.urlparse(url).path)[1]
            ext = os.path.splitext(msg.filename)[1]
            response = requests.get(url)
            msg.mime = response.headers['content-type']
            msg.file = NamedTemporaryFile(suffix=ext)
            msg.file.write(response.content)
            msg.file.seek(0)
            msg.path = msg.file.name
        elif attachment_type == '__Link':
            msg.type = MsgType.Link
            link_information = get_value(attachment, ('mercury', 'extensible_attachment', 'story_attachment'), {})
            title = get_value(link_information, ('title_with_entities', 'text'), '')
            description = get_value(link_information, ('description', 'text'), '')
            source = get_value(link_information, ('source', 'text'), None)
            if source:
                description += " (via %s)" % source
            preview = get_value(link_information, ('media', 'playable_url'), None) if \
                get_value(link_information, ('media', 'is_playable'), False) else None
            preview = preview or get_value(link_information, ('media', 'image', 'uri'), None)
            url = link_information['media']['url']
            msg.attributes = EFBMsgLinkAttribute(title=title,
                                                 description=description,
                                                 image=self.process_url(preview),
                                                 url=self.process_url(url))
        elif attachment_type == 'MessageLocation':
            msg.type = MsgType.Location
            link_information = get_value(attachment, ('mercury', 'extensible_attachment', 'story_attachment'), {})
            title = get_value(link_information, ('title_with_entities', 'text'), '')
            description = get_value(link_information, ('description', 'text'), '')
            msg.text = '\n'.join([title, description])
            preview = get_value(link_information, ('media', 'image', 'uri'), None)
            latitude, longitude = map(float. re.search(r'markers=([\d.]+)%2C([\d.]+)', preview).groups())
            msg.attributes = EFBMsgLocationAttribute(
                latitude=latitude, longitude=longitude
            )
        else:
            msg.type = MsgType.Unsupported
            msg.text = "Message type unsupported.\n%s" % msg.text

    def send(self, *args, **kwargs):
        result = super().send(*args, **kwargs)
        if result.startswith('mid.$'):
            self.sent_messages.add(result)
        return result

    def send_audio(self, file_id, message=None, thread_id=None, thread_type=ThreadType.USER):
        thread_id, thread_type = self._getThread(thread_id, thread_type)
        data = self._getSendData(message=message, thread_id=thread_id, thread_type=thread_type)

        data['action_type'] = 'ma-type:user-generated-message'
        data['has_attachment'] = True
        data['audio_ids[0]'] = file_id

        result = self._doSendRequest(data)
        if result.startswith('mid.$'):
            self.sent_messages.add(result)
        return result

    def send_file(self, file_id, message=None, thread_id=None, thread_type=ThreadType.USER):
        thread_id, thread_type = self._getThread(thread_id, thread_type)
        data = self._getSendData(message=message, thread_id=thread_id, thread_type=thread_type)

        data['action_type'] = 'ma-type:user-generated-message'
        data['has_attachment'] = True
        data['file_ids[0]'] = file_id

        result = self._doSendRequest(data)
        if result.startswith('mid.$'):
            self.sent_messages.add(result)
        return result

    def send_video(self, file_id, message=None, thread_id=None, thread_type=ThreadType.USER):
        thread_id, thread_type = self._getThread(thread_id, thread_type)
        data = self._getSendData(message=message, thread_id=thread_id, thread_type=thread_type)

        data['action_type'] = 'ma-type:user-generated-message'
        data['has_attachment'] = True
        data['video_ids[0]'] = file_id

        result = self._doSendRequest(data)
        if result.startswith('mid.$'):
            self.sent_messages.add(result)
        return result

    # Triggers

    def onMessage(self, mid: str=None, author_id: str=None, message: str=None, message_object: Message=None,
                  thread_id: str=None, thread_type: str=ThreadType.USER, ts: str=None, metadata=None, msg=None):
        """
        Called when the client is listening, and somebody sends a message

        Args:
            mid: The message ID
            author_id: The ID of the author
            message: (deprecated. Use `message_object.text` instead)
            message_object (models.Message): The message (As a `Message` object)
            thread_id: Thread ID that the message was sent to. See :ref:`intro_threads`
            thread_type (models.ThreadType): Type of thread that the message was sent to. See :ref:`intro_threads`
            ts: The timestamp of the message
            metadata: Extra metadata about the message
            msg: A full set of the data received
        """

        # Ignore messages sent by EFMS
        if mid in self.sent_messages:
            self.sent_messages.remove(mid)
            return

        self.logger.debug("[%s] Received message from Messenger: %s", mid, message_object)

        efb_msg = EFBMsg()
        efb_msg.uid = mid
        efb_msg.text = message_object.text
        efb_msg.type = MsgType.Text
        efb_msg.deliver_to = coordinator.master

        # Authors
        efb_msg.author = EFMSChat(self.channel, uid=author_id)
        efb_msg.chat = EFMSChat(self.channel, uid=thread_id)

        if message_object.mentions:
            mentions = dict()
            for i in message_object.mentions:
                mentions[(i.offset, i.offset + i.length)] = EFMSChat(self.channel, uid=i.thread_id)

        if message_object.emoji_size:
            efb_msg.text += " (%s)" % message_object.emoji_size.name[0]
            # " (S)", " (M)", " (L)"

        attachments = msg.get('delta', {}).get('attachments', [])

        if len(attachments) > 1:
            self.logger.debug("[%s] Multiple attachments detected. Splitting into %s messages.",
                              msg.uid, len(attachments))
            for idx, i in enumerate(attachments):
                sub_msg = copy.copy(efb_msg)
                sub_msg.uid += ".%d" % idx
                self.attach_media(sub_msg, i)
                coordinator.send_message(efb_msg)
            return

        if attachments:
            self.attach_media(efb_msg, attachments[0])

        coordinator.send_message(efb_msg)

        self.markAsDelivered(mid, thread_id)



from typing import TYPE_CHECKING

from fbchat import ThreadType

from ehforwarderbot.utils import extra

from .efms_chat import EFMSChat

if TYPE_CHECKING:
    from . import FBMessengerChannel


def catch_exceptions(func):
    """
    Decorator to wrap extra functions.
    Return the exception string when caught.
    """
    def catch_exceptions_wrap(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except Exception as e:
            return "Error occurred in %s(%s, %s): %r" % (func.__name__, args, kwargs, e)
    return catch_exceptions_wrap


class ExtraFunctionsManager:
    def __init__(self, channel: 'FBMessengerChannel'):
        self.channel = channel
        self.client = channel.client

    @catch_exceptions
    def threads_list(self, args: str) -> str:
        chats = self.channel.get_chats()
        msg = "You have %s thread(s) in your thread list.\n" % len(chats)
        for i in chats:
            msg += "\n{chat.chat_uid}: {chat.chat_name} [{type}]" \
                .format(chat=i,
                        type=i.vendor_specific.get('chat_type'))
        return msg

    @catch_exceptions
    def search_users(self, args: str) -> str:
        users = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForUsers(args, limit=10)))
        msg = "Found %s user(s).\n" % len(users)
        for i in users:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_groups(self, args: str) -> str:
        groups = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForGroups(args, limit=10)))
        msg = "Found %s group(s).\n" % len(groups)
        for i in groups:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_pages(self, args: str) -> str:
        pages = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForPages(args, limit=10)))
        msg = "Found %s page(s).\n" % len(pages)
        for i in pages:
            msg += "\n{chat.chat_uid}: {chat.chat_name}".format(chat=i)
        return msg

    @catch_exceptions
    def search_threads(self, args: str) -> str:
        threads = list(map(lambda a: EFMSChat(self.channel, a), self.client.searchForThreads(args, limit=10)))
        msg = "Found %s thread(s).\n" % len(threads)
        for i in threads:
            msg += "\n{chat.chat_uid}: {chat.chat_name} [{type}]" \
                .format(chat=i,
                        type=i.vendor_specific.get('chat_type'))
        return msg

    @catch_exceptions
    def add_to_group(self, args: str) -> str:
        """GroupID UserID [UserID ...]"""
        args = args.split(' ')
        if len(args) < 2:
            return "Group ID and user IDs are required"
        self.client.addUsersToGroup(args[1:], args[0])
        return "Users %s are successfully added to group %s." % (args[1:], args[0])

    @catch_exceptions
    def remove_from_group(self, args: str) -> str:
        """GroupID UserID"""
        args = args.split(' ')
        if len(args) != 2:
            return "Group ID and user ID are required"
        self.client.removeUserFromGroup(args[1], args[0])
        return "User %s is successfully removed from group %s." % (args[1], args[0])

    @catch_exceptions
    def set_nickname(self, args: str) -> str:
        """UserID nickname"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return "User ID and nickname are required."
        self.client.changeNickname(args[1], args[0], args[0])
        return "Nickname of %s is set to %s." % args

    @catch_exceptions
    def set_group_title(self, args: str) -> str:
        """GroupID title"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return "User ID and title are required."
        self.client.changeThreadTitle(args[1], args[0], thread_type=ThreadType.GROUP)
        return "Title of group %s is set to %s." % args

    @catch_exceptions
    def set_chat_emoji(self, args: str) -> str:
        """ChatID emoji"""
        args = args.split(' ', 1)
        if len(args) < 2:
            return "User ID and emoji are required."
        self.client.changeThreadEmoji(args[1], args[0])
        return "Emoji of group %s is set to %s." % args

    @catch_exceptions
    def set_member_nickname(self, args: str) -> str:
        """GroupID MemberID nickname"""
        args = args.split(' ', 2)
        if len(args) < 3:
            return "Group ID, member ID and are nickname required."
        self.client.changeNickname(args[2], args[1], args[0])
        return "Nickname of %s -> %s is set as %s." % args



"""
Authorize user and store session into the correct path.
"""

import getpass
import argparse
import pickle
import os

from fbchat import Client
from ehforwarderbot import coordinator, utils

from .__version__ import __version__
from . import FBMessengerChannel

def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("-p", "--profile", help="Choose a profile to start with. ", default="default")

    args = parser.parse_args()
    profile = args.profile
    coordinator.profile = profile
    path = utils.get_data_path(FBMessengerChannel.channel_id)

    print("EFB Facebook Messenger Slave Session Updater")
    print("============================================")
    print()
    print("You are running EFMS %s." % __version__)
    print()
    print("This will update and overwrite your EFMS token by")
    print("log into your account again manually.")
    print()
    print("You usually need to do this when you want to log into")
    print("a new account, or when the previous session is expired.")
    print()
    print("This session is written to EFB profile \"%s\"," % profile)
    print("which has its EFMS data located at")
    print(path)
    print()
    print("To continue, press Enter/Return.")

    input()

    email = input("Email: ")
    password = getpass.getpass("Password: ")
    client = Client(email, password)

    if not client.isLoggedIn():
        print("Log in failed. Please check the information above and try again.")
        exit(1)

    session_path = os.path.join(path, "session.pickle")

    with open(session_path, 'wb') as f:
        pickle.dump(client.getSession(), f)

    print("Your session has been successfully updated. It's stored at")
    print(session_path)
    print("Your session cookies is as valuable as your account credential.")
    print("Please keep them with equal care.")

if __name__ == '__main__':
    main()